// LED Beacon controller by Rob Swinbank @M0DTS Jan 2022
// Modulates an LED using the timepulse output of a NEO-6 GPS module, for good frequency stability.
// LED power is controllable in a few steps, Beacon 'on' between sepcified time period.
// 
//
//

#include <NeoSWSerial.h>
#include "LowPower.h"


NeoSWSerial gpsSerial(8, 9);

struct timeperiod
{
  float on;
  float off;
}; 

timeperiod timers[1];


//misc registers for configuring GPS
uint8_t GPSoff[] = {0xB5, 0x62, 0x06, 0x04, 0x04, 0x00, 0x00, 0x00,0x08, 0x00, 0x16, 0x74};
uint8_t GPSon[] = {0xB5, 0x62, 0x06, 0x04, 0x04, 0x00, 0x00, 0x00,0x09, 0x00, 0x17, 0x76};
uint8_t GPSeco[] = {0xB5, 0x62, 0x06, 0x11, 0x02, 0x00, 0x08, 0x04, 0x25, 0x95};
uint8_t GPSmax[] = {0xB5, 0x62, 0x06, 0x11, 0x02, 0x00, 0x08, 0x00, 0x21, 0x91};

//variables
bool satfix=false;
char line[80];
int hour=20;
int minute=55;
int second=0;

//time in floating point, eg 19.5 for 19:30
float hour_float=0;

long int gpsloop =0;
int pos=0;
bool GPSEnabled=true; //start up with GPS enabled


//used for calculating checksum of GPS commands
uint16_t calc_chksum( uint8_t *data, int count )
{
   uint8_t sum1 = 0;
   uint8_t sum2 = 0;
   int index;
   for ( index = 0; index < count; ++index )
   {
      sum1 += data[index];
      sum2 += sum1;
   }
   return (sum2 << 8) | sum1;
}



void setup() {
  //pin config
  pinMode(A5,OUTPUT);
  pinMode(3,OUTPUT);
  pinMode(4,OUTPUT);
  pinMode(5,OUTPUT);

  //initial pin state
  digitalWrite(A5,1);     //power switch hold
  digitalWrite(3,0);      //led power setting off
  digitalWrite(4,0);      //led power setting off
  digitalWrite(5,0);      //led power setting off
  

  gpsSerial.begin(9600);  //start gps software serial port
  Serial.begin(9600);     //start debug serial port
  gps_power(true);        //set gps RF section ot 'ON' state
  gps_eco(true);          //set GPS to econoy mode... not work?

  delay(500);             //wait for GPS power up.
  setfreq(1489);          //set GPS timepulse output to 1480Hz

  //beacon On/Off timer
  timers[0].on=16.2;
  timers[0].off=0;
  
}






void gps_power(bool up){
  if (up){
    for(int i=0;i<sizeof(GPSon);i++){
      gpsSerial.write(GPSon[i]);
    }
  }else{
    for(int i=0;i<sizeof(GPSoff);i++){
      gpsSerial.write(GPSoff[i]);
    }
  }
}

void gps_eco(bool eco){
    if (eco){
    for(int i=0;i<sizeof(GPSeco);i++){
      gpsSerial.write(GPSmax[i]);
    }
  }else{
    for(int i=0;i<sizeof(GPSmax);i++){
      gpsSerial.write(GPSmax[i]);
    }
  }
}

void setfreq(long int freq){
  //Sets both the Unlocked and Locked TimePulse output of the NEO-6 GPS Module (and others i assume) to the sepcified frequency.
  //Setting the unlocked value means that the signal is there before a valid GPS fix is available....
  //you could comment out the unlocked section if you don't want this to happen.

  byte cmd[40]={0xb5,0x62,0x06,0x31,0x20,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xb9,0x05,0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x00,0x6f,0x00,0x00,0x00,0x42,0x25};   //1465Hz  reference
  
  //Calculate frequency bytes and update command string plus checksum.
  //Locked freq
  cmd[18]=freq & 0xFF;
  cmd[19]=freq >> 8 &0xFF;
  cmd[20]=freq >> 16 &0xFF;
  
  //Unlocked freq = same as locked
  cmd[14]=cmd[18];
  cmd[15]=cmd[19];
  cmd[16]=cmd[20];
  
  uint16_t chk = calc_chksum(cmd+2,36);
  cmd[38]=chk & 0xFF;
  cmd[39]=chk >>8 & 0xFF;

  //Send frequency output command to ublox GPS module
  for(int i=0;i<sizeof(cmd);i++){
    gpsSerial.write(cmd[i]);
  }

  Serial.println("Started.");
}




void process_line(){
  int comma=0;

  if (line[1]=='G' & line[2]=='P' &line[3]=='G' &line[4]=='G' &line[5]=='A'){
    for (int i=0;i<80;i++){
      
      Serial.print(line[i]);      //debug output to pc serial port

      //get satellite 'fix' status from GPGGA sentence
      if(line[i]==','){
        comma++;
        if (comma==6){
          if (line[i+1]=='1'){
            satfix = true;
          }else{
            satfix = false;
          }
        }
      }
      //find end of sentence
      if(line[i]=='*'){
        break;
      }
    }
    
    Serial.println();


    //if we have satellite fix the get the time to some variables
    //also setup LED power level based on time
    //
    //wait until we have some good fix data (>1 sentences). Upon RF section power down/up we sometimes get invalid data on first sentence.
    //
    if(satfix & gpsloop >6){
      
      char bytes[2];
      bytes[0]=line[7];
      bytes[1]=line[8];
      String h = String((char *)bytes);
      hour = h.toInt();
      
      bytes[0]=line[9];
      bytes[1]=line[10];
      String m = String((char *)bytes);
      minute = m.toInt();
      
      bytes[0]=line[11];
      bytes[1]=line[12];
      String s = String((char *)bytes);
      second = s.toInt();
      hour_float = hour + (minute/60.0) +(second/60.0/60.0);

      
      if (hour_float>timers[0].on+1.0){   //1st hour on high power.
        
        //LED power ptions..... 0=100mA, 20=50mA, 40=12mA
        if (minute>=0 &minute<20){
          digitalWrite(3,0);
          digitalWrite(4,0);
          digitalWrite(5,1);
        }
        if (minute>=20 &minute<40){
          digitalWrite(3,0);
          digitalWrite(4,1);
          digitalWrite(5,0);
        }
        if (minute>=40 &minute<60){
          digitalWrite(3,1);
          digitalWrite(4,0);
          digitalWrite(5,0);
        }

      }else{
          digitalWrite(3,0);
          digitalWrite(4,0);
          digitalWrite(5,1);
      }
    }
  }
  gpsloop+=1;
}


void powerdown_fivemin(){
    for(int i=0;i<50;i++){    //5min=50
      LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
      delay(50);
      Serial.print((i+1)*8);
      Serial.println("s");
      delay(50);
    }
}

void powerdown_hour(){
    for(int i=0;i<412;i++){    //1h=3600s but 412*8 is about right.
      LowPower.powerDown(SLEEP_8S, ADC_OFF, BOD_OFF);
      delay(50);
      Serial.print((i+1)*8);
      Serial.println("s");
      delay(50);
    }
}







void loop() {

  //read battery volts from adc
  int adc = analogRead(0);
  float volts = ((adc/1024.0)*5.1)*2.0;
  //Serial.println(volts);

  //power off completely if battery volts are too low, i'm using a LiPo battery so this is necessary.
  if(volts <6.5){
    digitalWrite(A5,0);
  }

  //if GPS is set to run then process the data received.
  if(GPSEnabled){
    while (gpsSerial.available())
    {
      char c = gpsSerial.read();
      //get start of sentence, process previous line of data
      if (c=='$'){
       process_line();
        pos=0;
      }
      line[pos]=c;
      pos+=1;
      if (pos>79){pos=79;}
    }
    
    if (satfix & gpsloop >6){  
      //led for gps fix status - on
      digitalWrite(13,1); 
      if (hour_float>=timers[0].on | hour_float<timers[0].off ){

        //beacon is on
      }else{
        //go to sleep for a while if not in beacon 'on' time period
        GPSEnabled=false;
      }
    }else{
      //led for gps fix status - off
      digitalWrite(13,0);
    }
    
  }else{

    //this section does the 'sleeping', whole board then only uses ~11mA
    //could do better!!!
    //TODO
    //
    satfix=false;
    gps_power(false); //tell GPS to power RF section down
    delay(100);
    gpsSerial.end();    //stop serial interrupts waking from sleep
    Serial.print("Powering Down GPS...");
    

    //LED power drive off
    digitalWrite(3,0);
    digitalWrite(4,0);
    digitalWrite(5,0);
    digitalWrite(13,0);

    if (hour_float+1>timers[0].on){
      Serial.println("5min");
      delay(50);
      powerdown_fivemin();
    }else{
      Serial.println("1h");
      delay(50);
      powerdown_hour();
    }

    Serial.println("Powering Up GPS...");
    gpsloop=0;
    gpsSerial.begin(9600);    //restart gps serial port
    delay(50);
    gps_power(true);    //tell GPS to power RF section back on
    GPSEnabled=true;
  }

}
